/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "ns3/core-module.h"
#include "ns3/helper-module.h"
#include "ns3/mobility-module.h"
#include "ns3/process-module.h"
#include "ns3/simulator-module.h"
#include "ns3/global-route-manager.h"
#include "ns3/node-module.h"
#include "ns3/log.h"
#include "ns3/wifi-module.h"
#include "ns3/ipv4-l3-protocol.h"
#include "ns3/ipv4-global-routing-helper.h"
#include "ns3/ipv6-l3-protocol.h"
#include "ns3/ipv6-header.h"
#include "ns3/icmpv6-header.h"
#include "ns3/ipv6-static-routing.h"
#include "ns3/udp-header.h"
#include "ns3/realtime-simulator-impl.h"
#include "ns3/mpi-interface.h"

#include <sys/resource.h>

#ifdef NS3_MPI
#include <mpi.h>
#endif
using namespace ns3;

NS_LOG_COMPONENT_DEFINE ("QuaggaRocketfuelMpi");

#ifdef NS3_MPI
// Parameters
uint32_t nNodes = 2;
uint32_t stopTime = 60;


static void
SinkRx (Ptr<const Packet> p, const Address &ad)
{
  Ipv4Header ipv4;
  p->PeekHeader (ipv4);
  std::cout << "TTL: " << (unsigned)ipv4.GetTtl () << std::endl;
}

static void
SetRlimit ()
{
  int ret;
  struct rlimit limit;
  limit.rlim_cur = 1000000;
  limit.rlim_max = 1000000;

  ret = setrlimit(RLIMIT_NOFILE, &limit);
  if (ret == -1)
    {
       perror ("setrlimit");
    }
  return;
}
#endif

int
main (int argc, char *argv[])
{
  LogComponentEnable ("QuaggaRocketfuelMpi", LOG_LEVEL_INFO);
#ifdef NS3_MPI
  // Distributed simulation setup
  MpiInterface::Enable (&argc, &argv);
  GlobalValue::Bind ("SimulatorImplementationType",
                     StringValue ("ns3::DistributedSimulatorImpl"));

  uint32_t systemId = MpiInterface::GetSystemId ();
  uint32_t systemCount = MpiInterface::GetSize ();

  // Check for valid distributed parameters.
  // Must have 2 and only 2 Logical Processors (LPs)
  if (systemCount != 2)
    {
      std::cout << "This simulation requires 2 and only 2 logical processors." << std::endl;
      return 1;
    }

  CommandLine cmd;
  cmd.AddValue ("nNodes", "Number of Router nodes", nNodes);
  cmd.AddValue ("stopTime", "Time to stop(seconds)", stopTime);
  cmd.Parse (argc,argv);

  SetRlimit ();

  // 
  // Step o
  // Read Topology information
  // ------------------------------------------------------------
  // -- Read topology data.
  // --------------------------------------------

  // Pick a topology reader based in the requested format.

  Ptr<TopologyReader> inFile = 0;
  TopologyReaderHelper topoHelp;
  
  NodeContainer nodes;
  
  string format ("Rocketfuel");
  string input ("examples/topology-read/RocketFuel_toposample_1239_weights.txt");

  topoHelp.SetFileName(input);
  topoHelp.SetFileType(format);
  inFile = topoHelp.GetTopologyReader();

  if (inFile != 0)
    {
      nodes = inFile->Read ();
    }

  if (nodes.GetN () == 0)
    {
      NS_LOG_ERROR ("Problems reading node information the topology file. Failing.");
      return -1;
    }
  if (inFile->LinksSize () == 0)
    {
      NS_LOG_ERROR ("Problems reading the topology file. Failing.");
      return -1;
    }
  NS_LOG_INFO ("Rocketfuel topology created with " << nodes.GetN () << " nodes and " << 
               inFile->LinksSize () << " links (from " << input << ")");

  // 
  //  Step 1
  //  Node Basic Configuration
  // 

  // Internet stack install
  InternetStackHelper stack;    // IPv4 is required for GlobalRouteMan
  stack.Install (nodes);

  // 
  // Step 2
  // Address Configuration
  // 
  // 
  Ipv4AddressHelper ipv4AddrHelper;

  NS_LOG_INFO ("creating ip4 addresses");
  Ipv4AddressHelper address;
  address.SetBase ("10.0.0.0", "255.255.255.252");

  int totlinks = inFile->LinksSize ();
  NS_LOG_INFO ("creating node containers");
  NodeContainer nc[totlinks];
  TopologyReader::ConstLinksIterator iter;
  int i = 0;
  for ( iter = inFile->LinksBegin (); iter != inFile->LinksEnd (); iter++, i++ )
    {
      nc[i] = NodeContainer (iter->GetFromNode (), iter->GetToNode ());
    }


  NS_LOG_INFO ("creating net device containers");
  NetDeviceContainer ndc[totlinks];
  PointToPointHelper p2p;
  for (int i = 0; i < totlinks; i++)
    {
      // p2p.SetChannelAttribute ("Delay", TimeValue(MilliSeconds(weight[i])));
      p2p.SetChannelAttribute ("Delay", StringValue ("2ms"));
      p2p.SetDeviceAttribute ("DataRate", StringValue ("5Mbps"));
      ndc[i] = p2p.Install (nc[i]);
    }
//  p2p.EnablePcapAll ("quagga-rocketfuel");

  NS_LOG_INFO ("creating ipv4 interfaces");
  Ipv4InterfaceContainer ipic[totlinks];
  for (int i = 0; i < totlinks; i++)
    {
      ipic[i] = address.Assign (ndc[i]);
      address.NewNetwork ();
    }


  //  
  //  Application configuration for Nodes
  // 
  NS_LOG_INFO ("creating all the route");
  //  Ipv4GlobalRoutingHelper::PopulateRoutingTables ();

  NS_LOG_INFO ("creating quagga process");
  //run quagga programs in every node
  ProcessManagerHelper processManager;
  processManager.Install (nodes);
  QuaggaHelper quagga;
  quagga.EnableOspf (nodes);
  quagga.Install (nodes);  

  // 
  // Step 3
  // Traffic configuration
  // 
  UniformVariable unifRandom (0, nodes.GetN () - 1);
  unsigned int randomServerNumber = unifRandom.GetInteger (0, nodes.GetN () - 1);
  Ptr<Node> randomServerNode = nodes.Get (randomServerNumber);
  Ptr<Ipv4> ipv4Server = randomServerNode->GetObject<Ipv4> ();
  Ipv4InterfaceAddress iaddrServer = ipv4Server->GetAddress (1,0);
  Ipv4Address ipv4AddrServer = iaddrServer.GetLocal ();
  InetSocketAddress dst = InetSocketAddress ( ipv4AddrServer );
  if (systemId == 1)
    {
      // ------------------------------------------------------------
      // -- Send around packets to check the ttl
      // --------------------------------------------
      Config::SetDefault ("ns3::Ipv4RawSocketImpl::Protocol", StringValue ("2"));

      PacketSinkHelper sink = PacketSinkHelper ("ns3::Ipv4RawSocketFactory", dst);
      ApplicationContainer apps = sink.Install (randomServerNode);
      apps.Start (Seconds (0.0));
      apps.Stop (Seconds (3.0));

      // we trap the packet sink receiver to extract the TTL.
      Config::ConnectWithoutContext ("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx",
                                     MakeCallback (&SinkRx));
    }

  if (systemId == 0)
    {
      OnOffHelper onoff = OnOffHelper ("ns3::Ipv4RawSocketFactory", dst);
      onoff.SetAttribute ("OnTime", RandomVariableValue (ConstantVariable (1.0)));
      onoff.SetAttribute ("OffTime", RandomVariableValue (ConstantVariable (0.0)));
      onoff.SetAttribute ("DataRate", DataRateValue (DataRate (15000)));
      onoff.SetAttribute ("PacketSize", UintegerValue (1200));

      NodeContainer clientNodes;
      for ( unsigned int i = 0; i < nodes.GetN (); i++ )
        {
          if (i != randomServerNumber )
            {
              Ptr<Node> clientNode = nodes.Get (i);
              clientNodes.Add (clientNode);
            }
        }

      ApplicationContainer apps = onoff.Install (clientNodes);
      apps.Start (Seconds (1.0));
      apps.Stop (Seconds (2.0));
    }


  // 
  // Step 9
  // Now It's ready to GO!
  // 
  if (stopTime != 0)
    {
      Simulator::Stop (Seconds (stopTime));
    }
  Simulator::Run ();
  Simulator::Destroy ();


  std::cout << "End of experiment" << std::endl;

  return 0;
#else
  NS_FATAL_ERROR ("Can't use distributed simulator without MPI compiled in");
#endif
}
