/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "ns3/core-module.h"
#include "ns3/helper-module.h"
#include "ns3/mobility-module.h"
#include "ns3/process-module.h"
#include "ns3/simulator-module.h"
#include "ns3/global-route-manager.h"
#include "ns3/node-module.h"
#include "ns3/log.h"
#include "ns3/wifi-module.h"
#include "ns3/ipv4-l3-protocol.h"
#include "ns3/ipv4-global-routing-helper.h"
#include "ns3/ipv6-l3-protocol.h"
#include "ns3/ipv6-header.h"
#include "ns3/icmpv6-header.h"
#include "ns3/ipv6-static-routing.h"
#include "ns3/udp-header.h"
#include "ns3/realtime-simulator-impl.h"
#include <cstdio>
#include <cstdlib>
#include <string>
#include <fstream>
#include <sstream>
#include <iomanip>

#include "ns3/radvd.h"
#include "ns3/radvd-interface.h"
#include "ns3/radvd-prefix.h"

#if 0
#include "ns3/visualizer.h"
#include "ns3/pyviz.h"

#include "ns3/decorator-frontend.h"
#endif

using namespace ns3;

class BgpdConfig
{
private:
  static int index;
  uint32_t asn;
  std::string router_id;
  std::string loopback_addr;
  std::vector<std::string> *neighbors;
  std::map<std::string, uint32_t> *neighbor_asn;
  std::vector<std::string> *networks;
  bool isDefaultOriginate;
public:
  BgpdConfig()
  {
    asn = 7675+index;
    {
      std::stringstream ss;
      ss << "192.168.0." << index;
      router_id = ss.str();
    }
    {
      std::stringstream ss;
      ss << "2001:200:0:ffc4::" << index++;
      loopback_addr = ss.str();
    }
    neighbors = new std::vector<std::string>();
    neighbor_asn = new std::map<std::string, uint32_t>();
    networks = new std::vector<std::string>();
    isDefaultOriginate = false;
  }
  ~BgpdConfig()
  {
    delete neighbors;
    delete neighbor_asn;
    delete networks;
  }
  uint32_t getAsn()
  {
    return asn;
  }
  std::string getLoopbackAddr()
  {
    return loopback_addr;
  }
  void addNeighbor(std::string n, uint32_t asn)
  {
    neighbors->push_back(n);
    neighbor_asn->insert( std::map<std::string, uint32_t>::value_type( n, asn ));
  }
  void addNetwork(std::string n)
  {
    networks->push_back(n);
  }
  void defaultOriginate()
  {
    isDefaultOriginate = true;
  }
  void print(std::ostream& os) const
  {
    os << "hostname bgpd" << std::endl
       << "password zebra" << std::endl
       << "log stdout" << std::endl
       << "debug bgp" << std::endl
       << "debug bgp fsm" << std::endl
       << "debug bgp events" << std::endl
       << "debug bgp updates" << std::endl
       << "router bgp " << asn << std::endl
       << "  bgp router-id " << router_id << std::endl;    
    for( std::vector<std::string>::iterator it = neighbors->begin(); it != neighbors->end(); it++ )
      {
        os << "  neighbor " << *it << " remote-as " << (*neighbor_asn)[*it] << std::endl;
      }
    os << " address-family ipv4 unicast" << std::endl;
    // os << "  redistribute connected" << std::endl;
    // os << "  redistribute static" << std::endl;
    
    //   os << "  network 2001:200:800:/3" << std::endl;
    for( std::vector<std::string>::iterator it = neighbors->begin(); it != neighbors->end(); it++ )
      {
        os << "  neighbor " << *it << " activate" << std::endl;
        if( isDefaultOriginate == true )
          {
            os << "  neighbor " << *it << " default-originate" << std::endl;
          }
        //os << "  neighbor " << *it << " route-map set-nexthop out" << std::endl; 
        // os << "  neighbor " << *it << " next-hop-self" << std::endl; 
      }
    for( std::vector<std::string>::iterator it = networks->begin(); it != networks->end(); it++ )
      {
        os << "  network " << *it << std::endl;
      }
    os << " exit-address-family" << std::endl;
    os << "!" << std::endl;
    // os << "route-map set-nexthop permit 10" << std::endl;
    // os << " match ipv6 address all" << std::endl;
    //    os << " set ipv6 next-hop global 2001:200:0:8802::beaf" << std::endl;
    //    os << " set ipv6 next-hop local fe80::cbb5:591a
    // os << "log stdout" << std::endl;
  }
};

int BgpdConfig::index = 1;

std::ostream& operator << (std::ostream& os, BgpdConfig const& bgpd_config)
{
  bgpd_config.print (os);
  return os;
}

NS_LOG_COMPONENT_DEFINE ("camp-vm3");

// Parameters
uint32_t nrof_bgp_nodes = 10;
uint32_t stopTime = 120;
bool useTap = false;
bool useViz = false;

std::string expSquare = "200";     // meter
std::string outdir = "./";
int speed = 20;
std::string experiment = "manemo-rp-test";
std::string strategy = "none";
std::string input;
std::string runID;
std::string extDevIF = "eth0"; //Connection to RGNet
std::string apDevIF = "eth1"; //Connection to vm1 and vm2
int network = 0; // Network Selection (Default: RGNet 8880::/60)

void drawheader(int nodeid) {
  std::cout << "[Node " << nodeid << " Routing Table]" << std::endl;
  std::cout << "Destination";
  for(int i = 0; i < 32 - (int)strlen("Destination") + 8; i++) {
    std::cout << " ";
  }
  std::cout << "Prefix "; // 6 + 1
  std::cout << "Gateway";
  for(int i = 0; i < 32 - (int)strlen("Gateway") + 8; i++) {
    std::cout << " ";
  }
  std::cout << "IFID";
  std::cout << std::endl;
  for(int i = 0; i < 91; i++) {
    std::cout << "-";
  }
  std::cout << std::endl;
}

void AddRouteP2P (Ptr<Ipv6StaticRouting> sr, Ipv6Address dstnet, uint32_t dstprefix, Ipv6Address dstnext, uint32_t iface, uint32_t dstmetric=0) {
  int length1, length2;
  std::stringstream ss;
  ss << dstnet;
  length1 = (int)strlen(ss.str().c_str());
  ss.str("");
  ss << dstnext;
  length2 = (int)strlen(ss.str().c_str());
  sr->AddNetworkRouteTo(dstnet, Ipv6Prefix(dstprefix), dstnext, iface, dstmetric);
  std::cout << dstnet;
  for(int i = 0; i < 40 - length1; i++) {
    std::cout << " ";
  }
  std::cout << dstprefix;
  for(int i = 0; i < 5; i++) {
    std::cout << " ";
  }
  std::cout << dstnext << " ";
  for(int i = 0; i < 39 - length2; i++) {
    std::cout << " ";
  }
  std::cout << iface << std::endl;
}

void AddDefaultP2P (Ptr<Ipv6StaticRouting> sr, Ipv6Address defnext, uint32_t iface, Ipv6Address defprefix, uint32_t defmetric = 0) {
  if(network == 0) {
    sr->SetDefaultRoute(defnext, iface, defprefix);
  } else {
    sr->SetDefaultRoute(defnext, iface, defprefix);
  }
  std::cout << "Default Gateway";
  for(int i = 0; i < 40 - (int)strlen("Default Gateway"); i++) {
    std::cout << " ";
  }
  std::cout << "0";
  for(int i = 0; i < 6; i++) {
    std::cout << " ";
  }
  std::cout << defnext << " ";
  std::cout << iface << std::endl;
}

void 
CourseChange (std::string context, Ptr<const MobilityModel> model) 
{ 
  Vector position = model->GetPosition (); 
  NS_LOG_UNCOND (context << 
                 " x = " << position.x << ", y = " << position.y); 
} 

void
procExit (std::string path, uint16_t pid, int retcode)
{
  std::cout << "Process exit path=" << path << " pid=" << pid << " retcode=" << retcode << std::endl;
}

static void RunIp (Ptr<Node> node, Time at, std::string str)
{
  ProcessHelper process;
  ApplicationContainer apps;
  process.SetBinary ("build/debug/ip");
  process.SetStackSize (1<<16);
  process.ResetArguments();
  process.ParseArguments(str.c_str ());
  apps = process.Install (node);
  apps.Start (at);
}

static void AddAddress (Ptr<Node> node, Time at, const char *name, const char *address)
{
  std::ostringstream oss;
  oss << "-f inet addr add " << address << " dev " << name;
  RunIp (node, at, oss.str ());
}

int main (int argc, char *argv[]) {
  // 
  //  Step 0
  //  Node Basic Configuration
  // 

  // Enable Packet Pring
  ns3::PacketMetadata::Enable ();

  {
    std::stringstream sstr;
    sstr << "run-" << time(NULL);
    runID = sstr.str();

    sstr.clear();
    sstr << "BGP";
    strategy = sstr.str();

    sstr.clear();
    sstr << nrof_bgp_nodes;
    input = sstr.str();
  }

  CommandLine cmd;
  cmd.AddValue ("nrof_bgp_nodes", "Number of Router nodes", nrof_bgp_nodes);
  cmd.AddValue ("stopTime", "Time to stop(seconds)", stopTime);
  cmd.AddValue ("useViz", "Use Visualizer", useViz);
  cmd.AddValue("run", "Identifier for run.", runID);
  cmd.AddValue("outdir", "Directory for output file.", outdir);
  cmd.AddValue("network", "Network Configuration: 0 for RGNet(8880::/60), 1 for CampNet(ffc0::/60)", network);
  cmd.Parse (argc,argv);

  LogComponentEnable("manemo-exp", LOG_LEVEL_INFO);


  if (0)
    {
      Config::SetGlobal ("SimulatorImplementationType", StringValue ("ns3::RealtimeSimulatorImpl"));
      //      DynamicCast<RealtimeSimulatorImpl> (Simulator::GetImplementation ())->SetAttribute ("SynchronizationMode", EnumValue (RealtimeSimulatorImpl::SYNC_HARD_LIMIT));

      Config::SetGlobal ("ChecksumEnabled", BooleanValue (true));
    }


  // 
  //  Step 1
  //  Node Basic Configuration
  // 
  Ipv6AddressHelper ipv6AddrHelper;
  std::vector<BgpdConfig*> bgpd_configs;
  for(uint32_t i = 0; i < nrof_bgp_nodes; i++)
    {
      bgpd_configs.push_back(new BgpdConfig());
    }

  NodeContainer bgp_nodes;
  bgp_nodes.Create (nrof_bgp_nodes);
  NodeContainer tapHost;
  CsmaHelper csma;
  NetDeviceContainer tapDevs;


  // Address conf In virtual topology
  MobilityHelper mobility;
  Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
  positionAlloc->Add (Vector (0.0, 0.0, 0.0));
  positionAlloc->Add (Vector (5.0, 0.0, 0.0));
  mobility.SetPositionAllocator (positionAlloc);
  mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
  mobility.Install (bgp_nodes);

  WifiHelper wifi = WifiHelper::Default ();
  YansWifiPhyHelper phy = YansWifiPhyHelper::Default ();
  YansWifiChannelHelper phyChannel = YansWifiChannelHelper::Default ();
  NqosWifiMacHelper mac;
  mac.SetType ("ns3::AdhocWifiMac");
  wifi.SetStandard (WIFI_PHY_STANDARD_80211a);


  std::vector<uint32_t> graph;
#if 0
  graph.push_back(0);graph.push_back(1);
  graph.push_back(0);graph.push_back(2);
#endif 

#if 1
  graph.push_back(0);graph.push_back(7);

  graph.push_back(1);graph.push_back(2);
  graph.push_back(1);graph.push_back(7);

  graph.push_back(2);graph.push_back(3);

  graph.push_back(3);graph.push_back(4);
  graph.push_back(3);graph.push_back(7);

  graph.push_back(4);graph.push_back(9);

  graph.push_back(5);graph.push_back(6);
  graph.push_back(5);graph.push_back(7);

  graph.push_back(6);graph.push_back(8);

  graph.push_back(7);graph.push_back(8);

  graph.push_back(8);graph.push_back(9);
#endif

  //run quagga programs in every node
  ProcessManagerHelper processManager;
  processManager.SetLoader ("ns3::DlmLoaderFactory");
  processManager.SetNetworkStack("ns3::LinuxSocketFdFactory",
                                 "Library", StringValue ("libnet-next-2.6.so"));
  processManager.Install (bgp_nodes);

  std::vector<uint32_t>::iterator it = graph.begin();
  uint32_t link_count = 0;

  while( it != graph.end() )
    {
      uint32_t n1, n2;
      n1 = *it; it++;
      n2 = *it; it--;
      NodeContainer nodes_on_link;
      nodes_on_link = NodeContainer(bgp_nodes.Get(n1), bgp_nodes.Get(n2));

      NetDeviceContainer devices;
      phy.SetChannel (phyChannel.Create ());
      devices = wifi.Install (phy, mac, nodes_on_link);
      phy.EnablePcapAll ("bgp-example");

      std::stringstream v4network;

      std::stringstream n1_v4;
      std::stringstream n2_v4;
#if ns3
      v4network << "192.168." << (1+link_count) << ".0";
      // Internet stack install
      InternetStackHelper stack;    // IPv4 is required for GlobalRouteMan
      stack.Install (nodes_on_link);

      for (uint32_t i = 0; i < nodes_on_link; ++i) {
        // Set Flag not to accept default router from RA
        Ptr<Ipv6L3Protocol> ipv6 = bgp_nodes.Get(i)->GetObject<Ipv6L3Protocol> ();
        // TTL as 255
        ipv6->SetDefaultTtl(255);
      }

      Ipv4AddressHelper address;
      address.SetBase (v4network.str().c_str(), "255.255.255.0");
      Ipv4InterfaceContainer interfaces_v4 = address.Assign (devices);
      n1_v4 << interfaces_v4.GetAddress(0,0);
      n2_v4 << interfaces_v4.GetAddress(1,0);
#else  // linux-sock
      std::stringstream devstr, cmdstr;

      n1_v4 << "192.168." << (1+link_count) << ".1";
      v4network << n1_v4.str() << "/24";
      devstr << "sim" << devices.Get (0)->GetIfIndex ();
      AddAddress (bgp_nodes.Get (n1), Seconds (0.1), devstr.str ().c_str (), v4network.str ().c_str ());
      RunIp (bgp_nodes.Get (n1), Seconds (0.11), "link set lo up");
      cmdstr << "link set " << devstr.str().c_str() << " up";
      RunIp (bgp_nodes.Get (n1), Seconds (0.11), cmdstr.str());

      v4network.str ("");
      n2_v4 << "192.168." << (1+link_count) << ".2";
      v4network << n2_v4.str() << "/24";
      devstr.str ("");
      devstr << "sim" << devices.Get (1)->GetIfIndex ();
      AddAddress (bgp_nodes.Get (n2), Seconds (0.1), devstr.str().c_str(), v4network.str ().c_str ());
      RunIp (bgp_nodes.Get (n2), Seconds (0.11), "link set lo up");
      cmdstr.str("");
      cmdstr << "link set " << devstr.str().c_str() << " up";
      RunIp (bgp_nodes.Get (n2), Seconds (0.11), cmdstr.str());

      RunIp (bgp_nodes.Get (n1), Seconds (0.3), "route show table all");
      RunIp (bgp_nodes.Get (n1), Seconds (0.4), "addr list");
#endif

      bgpd_configs[n1]->addNeighbor(n2_v4.str(), bgpd_configs[n2]->getAsn());
      bgpd_configs[n2]->addNeighbor(n1_v4.str(), bgpd_configs[n1]->getAsn());

      /*
        Ptr<Ipv6L3Protocol> n1_l3 = bgp_nodes.Get(n1)->GetObject<Ipv6L3Protocol>();
        Ipv6StaticRoutingHelper n1_routing;
        Ptr<Ipv6StaticRouting> n1_sr = n1_routing.GetStaticRouting (n1_l3);
        n1_sr->AddNetworkRouteTo(Ipv6Address(bgpd_configs[n1]->getLoopbackAddr().c_str()), Ipv6Prefix(128),
        interfaces_v6.GetAddress(1,1), interfaces_v6.GetInterfaceIndex(0));

        Ptr<Ipv6L3Protocol> n2_l3 = bgp_nodes.Get(n2)->GetObject<Ipv6L3Protocol>();
        Ipv6StaticRoutingHelper n2_routing;
        Ptr<Ipv6StaticRouting> n2_sr = n2_routing.GetStaticRouting (n2_l3);
        n2_sr->AddNetworkRouteTo(Ipv6Address(bgpd_configs[n2]->getLoopbackAddr().c_str()), Ipv6Prefix(128),
        interfaces_v6.GetAddress(0,1), interfaces_v6.GetInterfaceIndex(1));
      */

      /*
        std::stringstream n1v6net;
        std::stringstream n2v6net;
        n1v6net << "200:200:0:" << std::setbase(16) << (0x8890 + link_count * 0x10 + n1) << "::/64";
        n2v6net << "200:200:0:" << std::setbase(16) << (0x8890 + link_count * 0x10 + n2) << "::/64";
        
        bgpd_configs[n1]->addNetwork(n1v6net.str());
        bgpd_configs[n2]->addNetwork(n2v6net.str());
      */

      it++;it++;
      link_count++;
    }

  bgpd_configs[9]->defaultOriginate();

  // bgpd config generation
  for(uint32_t i = 0; i < nrof_bgp_nodes; ++i)
    {
      std::stringstream conf_file;
      conf_file << "files-" << i << "/usr/local/etc/bgpd.conf";
      std::ofstream bgpdconf(conf_file.str().c_str());
      bgpdconf << *(bgpd_configs[i]);
      bgpdconf.close();
    }



	
  //  
  //  Application (zebra/bgpd) configuration for Nodes
  // 
  Ipv4GlobalRoutingHelper::PopulateRoutingTables ();

  ProcessHelper process;
  ProcessHelper process2;
  process.SetBinary ("zebra");
  process.SetStackSize (1<<20);
  process.AddArgument("-f");
  process.AddArgument("/usr/local/etc/zebra.conf");
  process2.SetBinary ("bgpd");
  process2.SetStackSize (1<<16);


  ApplicationContainer apps = process.Install (bgp_nodes);
  ApplicationContainer apps2 = process2.Install (bgp_nodes);

  for (uint32_t i = 0; i < nrof_bgp_nodes; ++i)
    {
      apps.Get(i)->SetStartTime (Seconds (1.0 + 1.01 * i));
      apps2.Get(i)->SetStartTime (Seconds (2.0 + 3.10 * i));
    }


  // Tap for external connectivity
  if (useTap)
    {
      tapDevs = csma.Install (bgp_nodes);
      tapHost.Create (1);
    }

  if (0)
    {
      std::string mode = "ConfigureLocal";
      std::string tapName = "mytap";

      Ipv4AddressHelper ipv4;
      ipv4.SetBase ("10.1.1.0", "255.255.255.0");
      Ipv4InterfaceContainer v4Tap = ipv4.Assign (tapDevs);
      ipv6AddrHelper.AssignWithoutAddress(tapDevs);

      TapBridgeHelper tapBridge (v4Tap.GetAddress (0));
      tapBridge.SetAttribute ("Mode", StringValue (mode));
      tapBridge.SetAttribute ("DeviceName", StringValue (tapName));
      tapBridge.Install (bgp_nodes.Get(2), tapDevs.Get(2));
    }

  // 
  // Output (generate file) section
  // 


  // 
  // Step 9
  // Decorator for movement trace
  // 
#if 0
  DecoratorFrontend decorator;
  //  decorator.SetArena(Rectangle(0, (width+1)*distance,
  //                               0, (height+1)*distance));
  decorator.SetArena(Rectangle(0, atof(expSquare.c_str()),
                               0, atof(expSquare.c_str())));
  std::ofstream out("decorator.log");
  decorator.SetStream(&out);

  decorator.Init();
#endif

  // 
  // Step 9
  // Now It's ready to GO!
  // 
  if (stopTime != 0)
    {
      Simulator::Stop (Seconds (stopTime));
    }
#if 0
  if (useViz)
    {
      Visualizer::Run ();
    }
  else
#endif
    {
      Simulator::Run ();
    }
  Simulator::Destroy ();


  std::cout << "End of experiment" << std::endl;

  return 0;
}
