/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 Hajime Tazaki
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Hajime Tazaki <tazaki@sfc.wide.ad.jp>
 */

#include "ns3/object-factory.h"
#include "quagga-helper.h"
#include "process-helper.h"
#include "ns3/names.h"
#include "ns3/ipv4-l3-protocol.h"
#include "ns3/utils.h"
#include <fstream>
#include <map>
#include <sys/stat.h>

namespace ns3 {

class QuaggaConfig: public Object
{
private:
  static int index;
  std::string router_id;
public:
  QuaggaConfig ()
  {
  }
  ~QuaggaConfig ()
  {
  }


  virtual void
  Print (std::ostream& os) const
  {
    os << "hostname zebra" << std::endl
       << "password zebra" << std::endl
       << "log stdout" << std::endl;
  }
};
std::ostream& operator << (std::ostream& os, QuaggaConfig const& config)
{
  config.Print (os);
  return os;
}

class OspfConfig : public QuaggaConfig
{
private:
  std::map<std::string, uint32_t> *networks;
public:
  OspfConfig ()
  {
    networks = new std::map<std::string, uint32_t> ();
    iflist = new std::vector<uint32_t> ();
  }
  ~OspfConfig ()
  {
    delete networks;
    delete iflist;
  }

  static TypeId 
  GetTypeId (void)
  {
    static TypeId tid = TypeId ("ns3::OspfConfig")
      .SetParent<QuaggaConfig> ()
      .AddConstructor<OspfConfig> ()
      ;
    return tid;
  }

  void
  addNetwork (Ipv4Address prefix, uint32_t plen, uint32_t area)
  {
    std::stringstream net;
    net << prefix << "/" << plen;
    networks->insert (std::map<std::string, uint32_t>::value_type (net.str (), area));
  }

  virtual void
  Print (std::ostream& os) const
  {
    QuaggaConfig::Print (os);
    //os << "log trap errors" << std::endl;
    os << "debug ospf event " << std::endl;
    //os << "debug ospf nsm " << std::endl;
    //     os << "debug ospf ism " << std::endl;
    // os << "debug ospf packet all " << std::endl;
    os << "!" << std::endl;

    for (std::vector<uint32_t>::iterator i = iflist->begin (); 
         i != iflist->end (); ++i)
      {
        os << "interface ns3-device" << (*i) << std::endl;
      }

    os << "router ospf " << std::endl;
    for (std::map<std::string, uint32_t>::iterator i = networks->begin (); 
         i != networks->end (); ++i)
      {
        os << "  network " << (*i).first << " area " << (*i).second << std::endl;
      }
    os << "!" << std::endl;
  }
  std::vector<uint32_t> *iflist;
};


QuaggaHelper::QuaggaHelper ()
  : m_ospfenable (false)
{
}

void 
QuaggaHelper::SetAttribute (std::string name, const AttributeValue &value)
{
}

// OSPF
void
QuaggaHelper::EnableOspf (NodeContainer nodes)
{
  //  for (uint32_t i = 0; i < nodes.GetN (); i++)
  for (uint32_t i = 0; i < nodes.GetN (); i ++)
    {
      Ptr<OspfConfig> ospf_conf = CreateObject<OspfConfig> ();
      nodes.Get (i)->AggregateObject(ospf_conf);

      for (uint32_t j = 0; j < nodes.Get (i)->GetNDevices (); j ++)
        {
          // Use primary address, if multiple
          //      Ipv4Address addr = interfaces->GetAddress (i, 0).GetLocal ();
          // uint32_t ifindex = interfaces.Get (i).second;
          // g_ospf_conf->iflist->insert (std::map<Ptr<Node>, uint32_t>::value_type (nodes.Get (i), ifindex));
        }
    }

  m_ospfenable = true;
  return;
}

void
QuaggaHelper::GenerateConfigZebra (Ptr<Node> node)
{
  QuaggaConfig *zebra_conf = new QuaggaConfig ();
  Ptr<Ipv4> ipv4 = node->GetObject<Ipv4> ();

  // config generation
  std::stringstream conf_dir, conf_file;
  // FIXME XXX
  conf_dir << "files-" << node->GetId () << "";
  ::mkdir (conf_dir.str ().c_str (), S_IRWXU | S_IRWXG);
  conf_dir << "/usr/";
  ::mkdir (conf_dir.str ().c_str (), S_IRWXU | S_IRWXG);
  conf_dir << "/local/";
  ::mkdir (conf_dir.str ().c_str (), S_IRWXU | S_IRWXG);
  conf_dir << "/etc/";
  ::mkdir (conf_dir.str ().c_str (), S_IRWXU | S_IRWXG);

  conf_file << conf_dir.str() << "/zebra.conf";
  std::ofstream conf;
  conf.open (conf_file.str ().c_str ());
  conf << *zebra_conf;
  conf << "debug zebra kernel" << std::endl;
  conf << "debug zebra events" << std::endl;
  conf << "debug zebra route" << std::endl;
  conf << "debug zebra packet" << std::endl;
  conf.close ();

}

void
QuaggaHelper::GenerateConfigOspf (Ptr<Node> node)
{
  Ptr<OspfConfig> ospf_conf = node->GetObject<OspfConfig>();
  Ptr<Ipv4> ipv4 = node->GetObject<Ipv4> ();

  for (uint32_t i = 0; i < ipv4->GetNInterfaces (); i ++)
    {
      Ipv4Address addr = ipv4->GetAddress (i, 0).GetLocal ();
      Ipv4Mask mask = ipv4->GetAddress (i, 0).GetMask ();
      Ipv4Address prefix = addr.CombineMask (mask);

      if (addr.IsEqual (Ipv4Address::GetLoopback()))
        {
          continue;
        }

      ospf_conf->iflist->push_back (i);
      ospf_conf->addNetwork (prefix, mask.GetPrefixLength (), 0);
    }

  // config generation
  std::stringstream conf_dir, conf_file;
  // FIXME XXX
  conf_dir << "files-" << node->GetId () << "";
  ::mkdir (conf_dir.str ().c_str (), S_IRWXU | S_IRWXG);
  conf_dir << "/usr/";
  ::mkdir (conf_dir.str ().c_str (), S_IRWXU | S_IRWXG);
  conf_dir << "/local/";
  ::mkdir (conf_dir.str ().c_str (), S_IRWXU | S_IRWXG);
  conf_dir << "/etc/";
  ::mkdir (conf_dir.str ().c_str (), S_IRWXU | S_IRWXG);

  conf_file << conf_dir.str() << "/ospfd.conf";
  std::ofstream conf;
  conf.open (conf_file.str ().c_str ());
  conf << *ospf_conf;
  conf.close ();

}


ApplicationContainer
QuaggaHelper::Install (Ptr<Node> node)
{
  return ApplicationContainer (InstallPriv (node));
}

ApplicationContainer
QuaggaHelper::Install (std::string nodeName)
{
  Ptr<Node> node = Names::Find<Node> (nodeName);
  return ApplicationContainer (InstallPriv (node));
}

ApplicationContainer
QuaggaHelper::Install (NodeContainer c)
{
  ApplicationContainer apps;
  for (NodeContainer::Iterator i = c.Begin (); i != c.End (); ++i)
    {
      apps.Add (InstallPriv (*i));
    }

  return apps;
}

ApplicationContainer
QuaggaHelper::InstallPriv (Ptr<Node> node) 
{
  ProcessHelper process;
  ApplicationContainer apps;

  GenerateConfigZebra (node);
  process.SetBinary ("build/debug/zebra");
  process.SetStackSize (1<<16);
  apps.Add (process.Install (node));
  apps.Get(0)->SetStartTime (Seconds (1.0 + 0.01 * node->GetId ()));
  node->AddApplication (apps.Get (0));

  // OSPF
  if (m_ospfenable)
    {
      GenerateConfigOspf (node);
      process.ResetArguments ();
      process.SetBinary ("build/debug/ospfd");
      apps.Add (process.Install (node));
      apps.Get(1)->SetStartTime (Seconds (1.0 + 0.1 * node->GetId ()));
      node->AddApplication (apps.Get (1));
    }

  return apps;
}

} // namespace ns3
