#include "copy-loader-factory.h"
#include "file-reader.h"
#include "elf-file-reader.h"
#include "file-writer.h"
#include "elf-file-writer.h"
#include "mmaped-file.h"
#include "ns3/log.h"
#include "ns3/assert.h"
#include "ns3/fatal-error.h"
#include "ns3/ptr.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <algorithm>
#include <dlfcn.h>
#include <errno.h>
#include <string.h>

NS_LOG_COMPONENT_DEFINE ("CopyLoaderFactory");

namespace ns3 {

class CopyLoader : public Loader
{
public:
  CopyLoader (uint32_t uid);
  virtual ~CopyLoader ();
  virtual void AddOverride (std::string from, std::string to);
  virtual void UnloadAll (void);
  virtual void *Load (std::string filename, int flag);
  virtual void Unload (void *module);
  virtual void *Lookup (void *module, std::string symbol);
private:
  struct FileToLoad
  {
    std::string filename;
    std::string dirname;
    std::string uniqueName;
    void *handle;
  };
  struct Override
  {
    std::string from;
    std::string to;
  };

  std::list<std::string> Split (std::string input, std::string sep) const;
  std::list<std::string> GetSearchDirectories (void) const;
  bool Exists (std::string filename) const;
  bool SearchFile (std::string filename, std::string *dirname) const;
  bool GatherLddDependencies (std::string filename);
  void EnsurePresent (std::string elfCache, struct FileToLoad fileToLoad);
  struct mdbg::ElfSectionHeader GetShtDynamic (mdbg::ElfFileReader reader);
  uint64_t GetDtStrTab (mdbg::ElfFileReader reader);
  std::string LookupUniqueDep (std::string filename);
  uint8_t NumberToChar (uint8_t c);
  std::string UidTo3LetterAscii (uint32_t uid);
  std::string AllocateUniqueName (std::string filename);
  std::string GetBasename (std::string filename);
  std::string GetDirname (std::string filename);
  uint64_t GetBase (mdbg::ElfFileReader elf);
  std::string GetLoaderCache (void);
  void AddFile (struct FileToLoad file);

  std::list<struct Override> m_overriden;
  std::list<struct FileToLoad> m_loaded;
  uint32_t m_uid;
};

CopyLoader::CopyLoader (uint32_t uid)
  : m_uid (uid)
{}
CopyLoader::~CopyLoader ()
{}

void 
CopyLoader::AddOverride (std::string from, std::string to)
{
  struct Override override;
  override.from = from;
  override.to = to;
  m_overriden.push_back (override);
}

void 
CopyLoader::UnloadAll (void)
{
  NS_LOG_FUNCTION (this);
  for (std::list<struct FileToLoad>::iterator i = m_loaded.begin (); i != m_loaded.end (); i++)
    {
      void *h = i->handle;
      if (h != 0)
	{
	  NS_LOG_DEBUG ("unload " << h);
	  ::dlclose (h);
	}
    }  
}

void *
CopyLoader::Load (std::string filename, int flag)
{
  NS_LOG_FUNCTION (this << filename << flag);
  // Note: ignore flag
  std::string dirname;
  if (!SearchFile (filename, &dirname))
    {
      NS_LOG_DEBUG ("could not find: " << filename);
      return false;
    }
  std::string fullname = dirname + "/" + filename;
  if (!GatherLddDependencies (fullname))
    {
      return 0;
    }
  // create the elf cache directory if needed.
  std::string elfLoaderCache = GetLoaderCache ();
  int retval = ::mkdir (elfLoaderCache.c_str (), S_IRWXU);
  if (retval == 0)
    {
      NS_LOG_DEBUG ("Created elf loader cache directory.");
    }
  for (std::list<struct FileToLoad>::iterator i = m_loaded.begin (); i != m_loaded.end (); i++)
    {
      // copy the file if not loaded yet.
      if (i->handle == 0)
	{
	  EnsurePresent (elfLoaderCache, *i);
	}
    }
  // finally, load the files.
  for (std::list<struct FileToLoad>::iterator i = m_loaded.begin (); i != m_loaded.end (); i++)
    {
      if (i->handle != 0)
	{
	  continue;
	}
      NS_LOG_DEBUG ("loading " << i->uniqueName.c_str ());
      ::dlerror ();
      void *handle = ::dlopen (i->uniqueName.c_str (), RTLD_LAZY | RTLD_LOCAL | RTLD_DEEPBIND);
      if (handle == 0)
	{
	  NS_FATAL_ERROR ("Could not open " << i->filename << " as " << i->uniqueName << 
			  " : " << ::dlerror ());
	}
      i->handle = handle;
      NS_LOG_DEBUG ("loaded " << i->uniqueName.c_str ());
    }
  for (std::list<struct FileToLoad>::iterator i = m_loaded.begin (); i != m_loaded.end (); i++)
    {
      if (i->filename == filename)
	{
	  return i->handle;
	}
    }
  NS_FATAL_ERROR ("can't reach here theoretically");
  return 0; // quiet compiler
}
void 
CopyLoader::Unload (void *module)
{
  // XXX
}
void *
CopyLoader::Lookup (void *module, std::string symbol)
{
  void * ptr = ::dlsym (module, symbol.c_str ());
  return ptr;
}

std::list<std::string>
CopyLoader::Split (std::string input, std::string sep) const
{
  NS_LOG_FUNCTION (this << input << sep);
  std::list<std::string> retval;
  std::string::size_type cur = 0, next;
  while (true)
    {
      next = input.find (sep, cur);
      if (next == cur)
	{
	  cur ++;
	  continue;
	} 
      else if (next == std::string::npos)
	{
	  if (input.size () != cur)
	    {
	      retval.push_back (input.substr (cur, input.size () - cur));
	    }
	  break;
	}
      retval.push_back (input.substr (cur, next - cur));
      cur = next + 1;
    }
  return retval;
}

std::list<std::string>
CopyLoader::GetSearchDirectories (void) const
{
  NS_LOG_FUNCTION (this);
  std::list<std::string> directories;
  char *ldLibraryPath = getenv ("LD_LIBRARY_PATH");
  if (ldLibraryPath != 0)
    {
      std::list<std::string> tmp = Split (ldLibraryPath, ":");
      directories.insert (directories.end (), 
			  tmp.begin (),
			  tmp.end ());
    }
  char *path = getenv ("PATH");
  if (path != 0)
    {
      std::list<std::string> tmp = Split (path, ":");
      directories.insert (directories.end (), 
			  tmp.begin (),
			  tmp.end ());
    }
  directories.push_back ("/lib");
  directories.push_back ("/usr/lib");
  directories.push_back (".");
  return directories;
}

bool
CopyLoader::Exists (std::string filename) const
{
  //NS_LOG_FUNCTION (this << filename);
  struct stat st;
  int retval = ::stat (filename.c_str (), &st);
  return retval == 0;
}

bool
CopyLoader::SearchFile (std::string filename, std::string *dirname) const
{
  NS_LOG_FUNCTION (this << filename);
  std::list<std::string> dirs = GetSearchDirectories ();
  for (std::list<std::string>::const_iterator i = dirs.begin (); i != dirs.end (); i++)
    {
      if (Exists (*i + "/" + filename))
	{
	  *dirname = *i;
	  NS_LOG_DEBUG ("Found: " << filename << " in " << *dirname);
	  return true;
	}
    }
  return false;
}

struct mdbg::ElfSectionHeader
CopyLoader::GetShtDynamic (mdbg::ElfFileReader elf)
{
  // search SHT_DYNAMIC
  struct mdbg::ElfHeader header = elf.GetHeader ();
  struct mdbg::ElfSectionHeader sectionHeader;
  NS_LOG_DEBUG ("nsections="<<header.e_shnum);
  for (uint64_t i = 0; i < header.e_shnum; i++)
    {
      sectionHeader = elf.ReadSectionHeader (i);
      //NS_LOG_DEBUG ("section=" << i << " sh_type=" << sectionHeader.sh_type<<
      //" sh_offset=" << sectionHeader.sh_offset);
      if (sectionHeader.sh_type == SHT_DYNAMIC)
	{
	  return sectionHeader;
	}
    }
  NS_ASSERT_MSG (false, "No SHT_DYNAMIC section found");
  return sectionHeader;
}
  
uint64_t 
CopyLoader::GetBase (mdbg::ElfFileReader elf)
{
  mdbg::ElfHeader header = elf.GetHeader ();
  uint64_t end = ~0; // the max value storable in a uint64_t
  uint64_t base = end;
  for (uint64_t i = 0; i < header.e_phnum; i++)
    {
      mdbg::ElfProgramHeader ph = elf.ReadProgramHeader (i);
      if (ph.p_type == PT_LOAD)
	{
	  uint64_t ph_base = ph.p_vaddr & ~(ph.p_align-1);
	  if (ph_base < base)
	    {
	      base = ph_base;
	    }
	}
    }
  if (base == end)
    {
      NS_LOG_ERROR ("Could not find base address.");
    }
  return base;
}

uint64_t 
CopyLoader::GetDtStrTab (mdbg::ElfFileReader elf)
{
  elf.Seek (GetShtDynamic (elf).sh_offset);
  uint64_t dt_strtab = 0;
  bool prelinked = false;
  while (true)
    {
      mdbg::ElfSxword d_tag = elf.ReadSxword ();
      mdbg::ElfXword d_val = elf.ReadXword ();
      //NS_LOG_DEBUG ("d_tag=" << d_tag << ", d_val=" << d_val);
      if (d_tag == DT_NULL)
	{
	  break;
	}
      else if (d_tag == DT_STRTAB)
	{
	  dt_strtab = d_val;
	}
      else if (d_tag == DT_GNU_PRELINKED)
	{
	  prelinked = true;
	}
    }
  NS_ASSERT_MSG (dt_strtab != 0, "Could not find DT_STRTAB");
  if (prelinked)
    {
      mdbg::ElfHeader header = elf.GetHeader ();
      NS_ASSERT (dt_strtab >= GetBase (elf));
      dt_strtab -= GetBase (elf);
    }
  NS_LOG_DEBUG ("DT_STRTAB="<<dt_strtab);
  return dt_strtab;
}

std::string
CopyLoader::GetBasename (std::string filename)
{
  std::string::size_type tmp = filename.find_last_of ("/");  
  if (tmp == std::string::npos)
    {
      return filename;
    }
  return filename.substr (tmp+1, filename.size () - (tmp+1));
}
std::string
CopyLoader::GetDirname (std::string filename)
{
  std::string::size_type tmp = filename.find_last_of ("/");  
  if (tmp == std::string::npos)
    {
      return "";
    }
  return filename.substr (0, tmp);
}

void
CopyLoader::AddFile (struct FileToLoad file)
{
  for (std::list<struct FileToLoad>::const_iterator i = m_loaded.begin (); i != m_loaded.end (); ++i)
    {
      if (i->dirname == file.dirname && i->filename == file.filename)
	{
	  return;
	}
    }
  file.handle = 0;
  m_loaded.push_back (file);
}
bool
CopyLoader::GatherLddDependencies (std::string fullname)
{
  /* We gather the dependencies for the input file using the 'ldd' program.
   * To do this, we do a standard fork+exec and forward the output of ldd
   * to the parent through a pipe which is then parsed in the parent.
   */
  struct FileToLoad file;
  file.dirname = GetDirname (fullname);
  file.filename = GetBasename (fullname);
  file.uniqueName = AllocateUniqueName (file.filename);
  AddFile (file);

  int pipefd[2];
  int retval = ::pipe (pipefd);
  if (retval == -1)
    {
      NS_LOG_ERROR ("Cannot create pipe: " << ::strerror (errno));
      return false;
    }
  pid_t pid = ::fork ();
  if (pid == 0)
    {
      // child.
      NS_LOG_DEBUG ("Child");
      ::close (pipefd[0]);
      int tmp;
      tmp = dup2 (pipefd[1], 1);
      if (tmp == -1)
	{
	  NS_LOG_ERROR ("Cannot redirect stdout");
	  ::exit (0);
	}
      retval = ::execl ("/usr/bin/ldd", "ldd", fullname.c_str (), (char *)NULL);
      if (retval == -1)
	{
	  NS_LOG_ERROR ("Cannot execl ldd: " << ::strerror (errno));
	}
      ::close (pipefd[1]);
      ::_exit (EXIT_SUCCESS);
    }
  else
    {
      // parent.
      NS_LOG_DEBUG ("Parent");
      ::close (pipefd[1]);
      std::string lddOutput;
      uint8_t c;
      ssize_t bytesRead = ::read (pipefd[0], &c, 1);
      while (bytesRead == 1)
	{
	  lddOutput.push_back (c);
	  bytesRead = ::read (pipefd[0], &c, 1);
	}
      ::close (pipefd[0]);
      std::string::size_type cur = 0;
      while (true)
	{
	  NS_LOG_DEBUG ("line=" << lddOutput);
	  std::string::size_type dep_start = lddOutput.find_first_not_of (" \t", cur);
	  std::string::size_type dep_end = lddOutput.find (" ", dep_start);
	  std::string::size_type full_start = lddOutput.find_first_of (">", dep_end);
	  full_start = lddOutput.find_first_not_of (" \t", full_start + 1);
	  std::string::size_type full_end = lddOutput.find_first_of (" \n", full_start);
	  NS_LOG_DEBUG ("dep_start=" << dep_start << " dep_end=" << dep_end << " full_start=" << full_start << " full_end=" << full_end);
	  if (dep_start != std::string::npos &&
	      full_start != std::string::npos &&
	      full_start > dep_start)
	    {
	      std::string depname = lddOutput.substr (dep_start, dep_end - dep_start);
	      std::string fulldepname = lddOutput.substr (full_start, full_end - (full_start));
	      NS_LOG_DEBUG (depname << "->" << fulldepname);
	      if (depname == "linux-gate.so.1" ||
		  depname == "/lib/ld-linux.so.2" ||
		  depname == "/lib64/ld-linux-x86-64.so.2" ||
		  depname == "/usr/lib/debug/ld-linux-x86-64.so.2" ||
		  depname == "linux-vdso.so.1")
		{
		  goto next;
		}
	      else
		{
		  struct FileToLoad dep;
		  dep.filename = depname;
		  dep.dirname = GetDirname (fulldepname);
		  dep.uniqueName = AllocateUniqueName (dep.filename);
		  AddFile (dep);
		}
	    }
	  else
	    {
	      break;
	    }
	next:
	  cur = lddOutput.find_first_of ("\n", full_start);
	  if (cur != std::string::npos)
	    {
	      cur++;
	    }
	  else
	    {
	      break;
	    }
	}
      
    }
  return true;
}
std::string 
CopyLoader::LookupUniqueDep (std::string filename)
{
  for (std::list<struct FileToLoad>::const_iterator i = m_loaded.begin (); i != m_loaded.end (); ++i)
    {
      if (filename == i->filename)
	{
	  return i->uniqueName;
	}
    }
  NS_ASSERT_MSG (false, "unable to find " << filename);
  return ""; // quiet compiler
}

void 
CopyLoader::EnsurePresent (std::string elfCache, struct FileToLoad fileToLoad)
{
  NS_LOG_FUNCTION (this << elfCache << fileToLoad.dirname << fileToLoad.filename << fileToLoad.uniqueName);
  std::string fromFile = fileToLoad.dirname + "/" + fileToLoad.filename;
  for (std::list<struct Override>::const_iterator i = m_overriden.begin (); 
       i != m_overriden.end (); ++i)
    {
      struct Override override = *i;
      if (override.from == fileToLoad.filename)
	{
	  std::string tmp;
	  if (!SearchFile (override.to, &tmp))
	    {
	      NS_LOG_ERROR ("Unable to find " + override.to );
	      return;
	    }
	  fromFile = tmp + "/" + override.to;
	  break;
	}
    }
  std::string toFile = elfCache + "/" + fileToLoad.uniqueName;

  ElfFileWriter writer (elfCache + "/" + fileToLoad.uniqueName);
  mdbg::MmapedFile file (fromFile);
  NS_ASSERT (file.IsOk ());
  mdbg::ElfFileReader elf = mdbg::ElfFileReader (file.GetBuffer (), file.GetSize ());

  struct mdbg::ElfHeader header = elf.GetHeader ();

  // copy everything
  writer.Write (file.GetBuffer (), file.GetSize ());

  // write the header to initialize the m_64 field properly.
  writer.Seek (0);
  writer.WriteHeader (header);

  // patch the section table to remove the VERSYM, VERNEED, and VERDEF headers
  writer.Seek (header.e_shoff);
  for (uint64_t i = 0 ; i < header.e_shnum; i++)
    {
      struct mdbg::ElfSectionHeader sectionHeader = elf.ReadSectionHeader (i);
      if (sectionHeader.sh_type == SHT_GNU_verdef ||
	  sectionHeader.sh_type == SHT_GNU_verneed ||
	  sectionHeader.sh_type == SHT_GNU_versym)
	{
	  NS_LOG_DEBUG ("clear section sh_type=" << std::hex << sectionHeader.sh_type);
	  sectionHeader.sh_type = SHT_NULL;
	}
      writer.WriteSectionHeader (sectionHeader);
    }

  struct mdbg::ElfSectionHeader shtDynamic = GetShtDynamic (elf);

  // Patch the DT_NEEDED, and, DT_SONAME entries
  elf.Seek (shtDynamic.sh_offset);
  struct mdbg::ElfFileReader str = elf;
  uint64_t dt_strtab = GetDtStrTab (elf);
  while (true)
    {
      mdbg::ElfSxword d_tag = elf.ReadSxword ();
      mdbg::ElfXword d_val = elf.ReadXword ();
      if (d_tag == DT_NEEDED)
	{
	  str.Seek (dt_strtab);
	  str.Skip (d_val);
	  std::string dep = str.ReadString ();
	  if (dep == "ld-linux.so.2" ||
	      dep == "ld-linux-x86-64.so.2")
	    {
	      continue;
	    }
	  std::string uniqueDep = LookupUniqueDep (dep);
	  NS_LOG_DEBUG ("needed=" << uniqueDep);
	  writer.Seek (dt_strtab);
	  writer.Skip (d_val);
	  writer.WriteString (uniqueDep);
	}
      else if (d_tag == DT_SONAME)
	{
	  writer.Seek (dt_strtab);
	  writer.Skip (d_val);
	  writer.WriteString (fileToLoad.uniqueName);
	}
      else if (d_tag == DT_NULL)
	{
	  break; // last entry.
	}
    }
}


std::string
CopyLoader::GetLoaderCache (void)
{
  return "elf-loader-cache";
}

uint8_t
CopyLoader::NumberToChar (uint8_t c)
{
  NS_ASSERT (c <= 60);
  if (c < 10)
    {
      return c + 0x30;
    }
  else if (c < (10+25))
    {
      return c - 10 + 0x41;
    }
  else if (c < 60)
    {
      return c - 10 - 25 + 0x61;
    }
  return 0; // quiet compiler
}
std::string 
CopyLoader::UidTo3LetterAscii (uint32_t uid)
{
  // we use base 60.
  if (uid >= 216000)
    {
      NS_FATAL_ERROR ("Please, report a bug: not enough unique strings for loaded code.");
    }
  uint32_t a = uid % 60;
  uint32_t b = (uid / 60) % 60;
  uint32_t c = (uid / 3600) % 60;
  std::string retval;
  retval.push_back (NumberToChar (c));
  retval.push_back (NumberToChar (b));
  retval.push_back (NumberToChar (a));
  return retval;
}
std::string 
CopyLoader::AllocateUniqueName (std::string filename)
{
  NS_ASSERT (filename.size () >= 3);
  std::string retval = UidTo3LetterAscii (m_uid) + filename.substr (3, filename.size () - 3);
  return retval;
}

NS_OBJECT_ENSURE_REGISTERED(CopyLoaderFactory);

TypeId 
CopyLoaderFactory::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::CopyLoaderFactory")
    .SetParent<LoaderFactory> ()
    .AddConstructor<CopyLoaderFactory> ()
    ;
  return tid;
}
CopyLoaderFactory::CopyLoaderFactory ()
{}
CopyLoaderFactory::~CopyLoaderFactory ()
{}
Loader *
CopyLoaderFactory::Create (int argc, char **argv, char **envp)
{
  // note: we ignore argc and co in this implementation.
  uint32_t uid = AllocateUid ();
  CopyLoader *loader = new CopyLoader (uid);
  return loader;
}
uint32_t
CopyLoaderFactory::AllocateUid (void)
{
  static uint32_t uid = 0;
  uid++;
  return uid;
}

} // namespace ns3
