#include "elf-file-reader.h"
#include "ns3/log.h"

NS_LOG_COMPONENT_DEFINE ("ElfFileReader");

#define ASCII_E 0x45
#define ASCII_L 0x4C
#define ASCII_F 0x46

using namespace ns3;

namespace mdbg {

ElfFileReader::ElfFileReader (const uint8_t *buffer, uint64_t size)
  : m_reader (buffer, size)
{
  Initialize ();
}

bool 
ElfFileReader::IsOk (void) const
{
  return m_reader.IsOk ();
}
uint64_t 
ElfFileReader::GetOffset (void) const
{
  return m_reader.GetOffset ();
}
void 
ElfFileReader::Seek (uint64_t offset)
{
  m_reader.Seek (offset);
}
void 
ElfFileReader::Skip (uint64_t offset)
{
  m_reader.Skip (offset);
}
void
ElfFileReader::Initialize (void)
{
  NS_LOG_FUNCTION (this);
  m_reader.Seek (0);
  uint32_t tmp = m_reader.ReadU8 ();
  if (tmp != 0x7f)
    {
      NS_LOG_DEBUG ("Invalid magic ELF number: " << tmp);
      m_reader.SetFail ();
      return;
    }
  tmp = m_reader.ReadU8 ();
  if (tmp != ASCII_E)
    {
      NS_LOG_DEBUG ("Invalid ELF E: " << tmp);
      m_reader.SetFail ();
      return;
    }
  tmp = m_reader.ReadU8 ();
  if (tmp != ASCII_L)
    {
      NS_LOG_DEBUG ("Invalid ELF L: " << tmp);
      m_reader.SetFail ();
      return;
    }
  tmp = m_reader.ReadU8 ();
  if (tmp != ASCII_F)
    {
      NS_LOG_DEBUG ("Invalid ELF F: " << tmp);
      m_reader.SetFail ();
      return;
    }
  m_header.ei_class = m_reader.ReadU8 ();
  if (m_header.ei_class == 1)
    {
      m_64 = false;
    }
  else if (m_header.ei_class == 2)
    {
      m_64 = true;
    }
  else
    {
      NS_LOG_DEBUG ("Invalid ELF ei_class: " << (uint32_t) m_header.ei_class);
      m_reader.SetFail ();
      return;
    }
  m_header.ei_data = m_reader.ReadU8 ();
  if (m_header.ei_data == 1)
    {
      m_reader.SetLsb ();
    }
  else if (m_header.ei_data == 2)
    {
      m_reader.SetMsb ();
    }
  else
    {
      NS_LOG_DEBUG ("Invalid ELF ei_data: " << (uint32_t) m_header.ei_data);
      m_reader.SetFail ();
      return;
    }
  m_header.ei_version = m_reader.ReadU8 ();
  if (m_header.ei_version != 1)
    {
      NS_LOG_DEBUG ("Invalid ELF ei_version: " << (uint32_t) m_header.ei_version);
      m_reader.SetFail ();
      return;
    }
  for (uint8_t i = 0; i < 16-7; i++)
    {
      uint8_t pad;
      pad = m_reader.ReadU8 ();
      // just ignore them.
    }
  m_header.e_type = ReadHalf ();
  m_header.e_machine = ReadHalf ();
  m_header.e_version = ReadWord ();
  m_header.e_entry = ReadAddr ();
  m_header.e_phoff = ReadOff ();
  m_header.e_shoff = ReadOff ();
  m_header.e_flags = ReadWord();
  m_header.e_ehsize = ReadHalf ();
  m_header.e_phentsize = ReadHalf ();
  m_header.e_phnum = ReadHalf ();
  m_header.e_shentsize = ReadHalf ();
  m_header.e_shnum = ReadHalf ();
  m_header.e_shstrndx = ReadHalf ();

  if (m_header.e_version != 1)
    {
      NS_LOG_DEBUG ("cannot load elf file version=" << m_header.e_version);
      m_reader.SetFail ();
    }
}

ElfHalf
ElfFileReader::ReadHalf (void)
{
  return m_reader.ReadU16 ();
}
ElfWord
ElfFileReader::ReadWord (void)
{
  return m_reader.ReadU32 ();
}
ElfXword
ElfFileReader::ReadXword (void)
{
  if (m_64)
    {
      return m_reader.ReadU64 ();
    }
  else
    {
      return m_reader.ReadU32 ();
    }
}
ElfSxword 
ElfFileReader::ReadSxword (void)
{
  if (m_64)
    {
      return m_reader.ReadU64 ();
    }
  else
    {
      return m_reader.ReadU32 ();
    }
}
ElfSword 
ElfFileReader::ReadSword (void)
{
  return m_reader.ReadU32 ();
}
ElfOff
ElfFileReader::ReadOff (void)
{
  if (m_64)
    {
      return m_reader.ReadU64 ();
    }
  else
    {
      return m_reader.ReadU32 ();
    }
}
ElfAddr
ElfFileReader::ReadAddr (void)
{
  if (m_64)
    {
      return m_reader.ReadU64 ();
    }
  else
    {
      return m_reader.ReadU32 ();
    }
}
std::string 
ElfFileReader::ReadString (void)
{
  std::string str;
  uint8_t tmp = m_reader.ReadU8 ();
  while (tmp != 0)
    {
      str.push_back (tmp);
      tmp = m_reader.ReadU8 ();
    }
  return str;
}

struct ElfHeader 
ElfFileReader::GetHeader (void)
{
  return m_header;
}

struct ElfSectionHeader 
ElfFileReader::ReadSectionHeader (uint64_t shn)
{
  NS_LOG_FUNCTION (this);
  if (shn >= m_header.e_shnum)
    {
      m_reader.SetFail ();
    }
  m_reader.Seek (m_header.e_shoff);
  m_reader.Skip (m_header.e_shentsize * shn);
  return ReadSectionHeader ();
}

struct ElfSectionHeader 
ElfFileReader::ReadSectionHeader (void)
{
  NS_LOG_FUNCTION (this << m_reader.GetOffset ());
  struct ElfSectionHeader header;
  header.sh_name = ReadWord ();
  header.sh_type = ReadWord ();
  header.sh_flags = ReadXword ();
  header.sh_addr = ReadAddr ();
  header.sh_offset = ReadOff ();
  header.sh_size = ReadXword ();
  header.sh_link = ReadWord ();
  header.sh_info = ReadWord ();
  header.sh_addralign = ReadXword ();
  header.sh_entsize = ReadXword ();

  return header;
}

struct ElfProgramHeader 
ElfFileReader::ReadProgramHeader (void)
{
  NS_LOG_FUNCTION (this);
  struct ElfProgramHeader header;
  header.p_type = ReadWord ();
  if (m_64)
    {
      header.p_flags = ReadWord ();
    }
  header.p_offset = ReadOff ();
  header.p_vaddr = ReadAddr ();
  header.p_paddr = ReadAddr ();
  header.p_filesz = ReadXword ();
  header.p_memsz = ReadXword ();
  header.p_align = ReadXword ();
  return header;
}
struct ElfProgramHeader 
ElfFileReader::ReadProgramHeader (uint64_t phn)
{
  struct ElfHeader header = GetHeader ();
  struct ElfProgramHeader ph;
  if (phn >= header.e_phnum)
    {
      m_reader.SetFail ();
      return ph;
    }
  m_reader.Seek (header.e_phoff);
  m_reader.Skip (header.e_phentsize * phn);
  ph = ReadProgramHeader ();
  return ph;
}


} // namespace mdbg
