#include "elf-file-writer.h"
#include "ns3/log.h"

NS_LOG_COMPONENT_DEFINE ("ElfFileWriter");

#define ASCII_E 0x45
#define ASCII_L 0x4C
#define ASCII_F 0x46

namespace ns3 {

ElfFileWriter::ElfFileWriter (std::string filename)
  : m_writer (filename)
{}

uint64_t 
ElfFileWriter::GetOffset (void)
{
  return m_writer.GetOffset ();
}
void 
ElfFileWriter::WriteHeader (struct mdbg::ElfHeader header)
{
  m_writer.WriteU8 (0x7f);
  m_writer.WriteU8 (ASCII_E);
  m_writer.WriteU8 (ASCII_L);
  m_writer.WriteU8 (ASCII_F);
  m_writer.WriteU8 (header.ei_class);
  m_writer.WriteU8 (header.ei_data);
  m_writer.WriteU8 (header.ei_version);
  if (header.ei_class == 1)
    {
      m_64 = false;
    }
  else if (header.ei_class == 2)
    {
      m_64 = true;
    }
  else
    {
      NS_LOG_DEBUG ("Invalid ei_class=" << (uint32_t) header.ei_class);
      m_writer.SetFail ();
      return;
    }
  if (header.ei_data == 1)
    {
      m_writer.SetLsb ();
    }
  else if (header.ei_data == 2)
    {
      m_writer.SetMsb ();
    }
  else
    {
      NS_LOG_DEBUG ("Invalid ELF ei_data: " << (uint32_t) header.ei_data);
      m_writer.SetFail ();
      return;
    }
  uint32_t padding = 16 - m_writer.GetOffset ();
  for (uint32_t i = 0; i < padding; i++)
    {
      m_writer.WriteU8 (0);
    }

  WriteHalf (header.e_type);
  WriteHalf (header.e_machine);
  WriteWord (header.e_version);
  WriteAddr (header.e_entry);
  WriteOff (header.e_phoff);
  WriteOff (header.e_shoff);
  WriteWord (header.e_flags);
  WriteHalf (header.e_ehsize);
  WriteHalf (header.e_phentsize);
  WriteHalf (header.e_phnum);
  WriteHalf (header.e_shentsize);
  WriteHalf (header.e_shnum);
  WriteHalf (header.e_shstrndx);
}

void 
ElfFileWriter::WriteSectionHeader (struct mdbg::ElfSectionHeader header)
{
  WriteWord (header.sh_name);
  WriteWord (header.sh_type);
  WriteXword (header.sh_flags);
  WriteAddr (header.sh_addr);
  WriteOff  (header.sh_offset);
  WriteXword (header.sh_size);
  WriteWord (header.sh_link);
  WriteWord (header.sh_info);
  WriteXword (header.sh_addralign);
  WriteXword (header.sh_entsize);
}
void 
ElfFileWriter::Seek (uint64_t offset)
{
  m_writer.Seek (offset);
}
void 
ElfFileWriter::Skip (int64_t offset)
{
  m_writer.Skip (offset);
}
void 
ElfFileWriter::WriteU8 (uint8_t v)
{
  m_writer.WriteU8 (v);
}
void 
ElfFileWriter::Write (uint8_t *buffer, uint64_t size)
{
  m_writer.Write (buffer, size);
}

void
ElfFileWriter::WriteHalf (mdbg::ElfHalf v)
{
  m_writer.WriteU16 (v);
}
void
ElfFileWriter::WriteXword (mdbg::ElfXword v)
{
  if (m_64)
    {
      m_writer.WriteU64 (v);
    }
  else
    {
      m_writer.WriteU32 (v);
    }
}
void 
ElfFileWriter::WriteSxword (mdbg::ElfSxword v)
{
  if (m_64)
    {
      m_writer.WriteU64 (v);
    }
  else
    {
      m_writer.WriteU32 (v);
    }
}
void
ElfFileWriter::WriteWord (mdbg::ElfWord v)
{
  m_writer.WriteU32 (v);
}
void
ElfFileWriter::WriteSword (mdbg::ElfSword v)
{
  m_writer.WriteU32 (v);
}
void
ElfFileWriter::WriteOff (mdbg::ElfOff v)
{
  if (m_64)
    {
      m_writer.WriteU64 (v);
    }
  else
    {
      m_writer.WriteU32 (v);
    }
}
void
ElfFileWriter::WriteAddr (mdbg::ElfAddr v)
{
  if (m_64)
    {
      m_writer.WriteU64 (v);
    }
  else
    {
      m_writer.WriteU32 (v);
    }
}
void 
ElfFileWriter::WriteString (std::string str)
{
  for (uint32_t i = 0; i < str.size (); i++)
    {
      WriteU8 (str[i]);
    }
  WriteU8 (0);
}

} // namespace ns3
