#include "file-reader.h"

namespace mdbg {

FileReader::FileReader (const uint8_t *buffer, uint64_t size)
  : m_buffer (buffer),
    m_size (size),
    m_current (0),
    m_lsb (false),
    m_ok (true)
{}
FileReader::~FileReader ()
{}
void 
FileReader::SetFail (void)
{
  m_ok = false;
}
bool 
FileReader::IsOk (void) const
{
  return m_ok;
}
void 
FileReader::SetMsb (void)
{
  m_lsb = false;
}
void 
FileReader::SetLsb (void)
{
  m_lsb = true;
}
bool
FileReader::HasLeft (uint8_t n) const
{
  return m_current <= m_size && m_size - m_current >= n;
}
uint8_t
FileReader::ReadU8 (void)
{
  if (!HasLeft (1))
    {
      m_ok = false;
      return 0xff;
    }
  uint8_t retval = m_buffer[m_current];
  m_current++;
  return retval;
}
uint16_t
FileReader::ReadU16 (void)
{
  if (!HasLeft (2))
    {
      m_ok = false;
      return 0xffff;
    }
  uint8_t byte0 = m_buffer[m_current];
  uint8_t byte1 = m_buffer[m_current+1];
  uint16_t retval;
  if (m_lsb) 
    {
      retval = byte0 + (byte1<<8);
    } 
  else 
    {
      retval = (byte0<<8) + byte1;
    }
  m_current += 2;
  return retval;
}
uint32_t
FileReader::ReadU32 (void)
{
  if (!HasLeft (4))
    {
      m_ok = false;
      return 0xffffffff;
    }
  uint8_t byte0 = m_buffer[m_current];
  uint8_t byte1 = m_buffer[m_current+1];
  uint8_t byte2 = m_buffer[m_current+2];
  uint8_t byte3 = m_buffer[m_current+3];
  uint32_t retval;
  if (m_lsb) 
    {
      retval = byte0 + (byte1<<8) + (byte2<<16) + (byte3<<24);
    } 
  else 
    {
      retval = (byte0<<24) + (byte1<<16) + (byte2<<8) + byte3;
    }
  m_current += 4;
  return retval;
}
uint64_t
FileReader::ReadU64 (void)
{
  if (!HasLeft (8))
    {
      m_ok = false;
      return 0xffffffffffffffffLL;
    }

  uint64_t byte0 = m_buffer[m_current];
  uint64_t byte1 = m_buffer[m_current+1];
  uint64_t byte2 = m_buffer[m_current+2];
  uint64_t byte3 = m_buffer[m_current+3];
  uint64_t byte4 = m_buffer[m_current+4];
  uint64_t byte5 = m_buffer[m_current+5];
  uint64_t byte6 = m_buffer[m_current+6];
  uint64_t byte7 = m_buffer[m_current+7];
  uint64_t retval;
  if (m_lsb) 
    {
      retval = byte0 + (byte1<<8) + (byte2<<16) + (byte3<<24) 
	+ (byte4<<32) + (byte5<<40) + (byte6<<48) + (byte7<<56);
    } 
  else 
    {
      retval = (byte0<<56) + (byte1<<48) + (byte2<<40) + (byte3<<32) 
	+ (byte4<<24) + (byte5<<16) + (byte6<<8) + (byte7<<0);
    }
  m_current += 8;
  return retval;
}
int8_t
FileReader::ReadS8 (void)
{
  return ReadU8 ();
}
int16_t
FileReader::ReadS16 (void)
{
  return ReadU16 ();
}
int32_t
FileReader::ReadS32 (void)
{
  return ReadU32 ();
}
int64_t
FileReader::ReadS64 (void)
{
  return ReadU64 ();
}
uint64_t 
FileReader::GetOffset (void) const
{
  return m_current;
}
void 
FileReader::Seek (uint64_t offset)
{
  m_current = offset;
}
void 
FileReader::Skip (uint64_t offset)
{
  m_current += offset;
}

} // namespace mdbg
