#include "file-writer.h"
#include "ns3/assert.h"
#include "ns3/log.h"
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

NS_LOG_COMPONENT_DEFINE ("FileWriter");

namespace ns3 {

FileWriter::FileWriter (std::string filename)
{
  m_fd = ::open (filename.c_str (), O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
  if (m_fd == -1)
    {
      m_ok = false;
      NS_LOG_DEBUG (filename << ": " << strerror (errno));
    }
}

FileWriter::~FileWriter ()
{
  int retval;
  retval = ::close (m_fd);
  NS_ASSERT (retval == 0);
}

bool 
FileWriter::IsOk (void) const
{
  return m_ok;
}
void 
FileWriter::SetFail (void)
{
  m_ok = false;
}

uint64_t 
FileWriter::GetOffset (void)
{
  off_t cur = ::lseek (m_fd, 0, SEEK_CUR);
  if (cur == -1)
    {
      m_ok = false;
    }
  return cur;
}

void 
FileWriter::SetLsb (void)
{
  m_lsb = true;
}
void 
FileWriter::SetMsb (void)
{
  m_lsb = false;
}
void 
FileWriter::Seek (uint64_t offset)
{
  off_t retval;
  retval = ::lseek (m_fd, offset, SEEK_SET);
  if (retval == -1)
    {
      m_ok = false;
    }
}
void 
FileWriter::Skip (int64_t offset)
{
  off_t retval;
  retval = ::lseek (m_fd, offset, SEEK_CUR);
  if (retval == -1)
    {
      m_ok = false;
    }
}
void 
FileWriter::WriteU8  (uint8_t v)
{
  int retval;
  retval = ::write (m_fd, &v, 1);
  if (retval == -1)
    {
      m_ok = false;
    }
}
void 
FileWriter::WriteU16 (uint16_t v)
{
  uint8_t byte0 = v & 0xff;
  uint8_t byte1 = (v >> 8) & 0xff;
  if (m_lsb)
    {
      WriteU8 (byte0);
      WriteU8 (byte1);
    }
  else
    {
      WriteU8 (byte1);
      WriteU8 (byte0);
    }
}
void 
FileWriter::WriteU32 (uint32_t v)
{
  uint8_t byte0 = v & 0xff;
  uint8_t byte1 = (v >> 8) & 0xff;
  uint8_t byte2 = (v >> 16) & 0xff;
  uint8_t byte3 = (v >> 24) & 0xff;
  if (m_lsb)
    {
      WriteU8 (byte0);
      WriteU8 (byte1);
      WriteU8 (byte2);
      WriteU8 (byte3);
    }
  else
    {
      WriteU8 (byte3);
      WriteU8 (byte2);
      WriteU8 (byte1);
      WriteU8 (byte0);
    }
}
void
FileWriter::WriteU64 (uint64_t v)
{
  uint8_t byte0 = v & 0xff;
  uint8_t byte1 = (v >> 8) & 0xff;
  uint8_t byte2 = (v >> 16) & 0xff;
  uint8_t byte3 = (v >> 24) & 0xff;
  uint8_t byte4 = (v >> 32) & 0xff;
  uint8_t byte5 = (v >> 40) & 0xff;
  uint8_t byte6 = (v >> 48) & 0xff;
  uint8_t byte7 = (v >> 54) & 0xff;
  if (m_lsb)
    {
      WriteU8 (byte0);
      WriteU8 (byte1);
      WriteU8 (byte2);
      WriteU8 (byte3);
      WriteU8 (byte4);
      WriteU8 (byte5);
      WriteU8 (byte6);
      WriteU8 (byte7);
    }
  else
    {
      WriteU8 (byte7);
      WriteU8 (byte6);
      WriteU8 (byte5);
      WriteU8 (byte4);
      WriteU8 (byte3);
      WriteU8 (byte2);
      WriteU8 (byte1);
      WriteU8 (byte0);
    }
}
void 
FileWriter::Write (uint8_t *buffer, uint64_t size)
{
  int retval;
  retval = ::write (m_fd, buffer, size);
  if (retval == -1)
    {
      m_ok = false;
    }
}

} // namespace ns3
