#include "linux-socket-fd-factory.h"
#include "linux-socket-fd.h"
#include "loader-factory.h"
#include "process-manager.h"
#include "process.h"
#include "utils.h"
#include "task-manager.h"
#include "sim/include/sim-init.h"
#include "ns3/log.h"
#include "ns3/string.h"
#include "ns3/node.h"
#include "ns3/net-device.h"
#include "ns3/random-variable.h"
#include "ns3/event-id.h"
#include "ns3/simulator.h"
#include "ns3/mac48-address.h"
#include "ns3/packet.h"
#include <dlfcn.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <arpa/inet.h>
#include <fcntl.h>

NS_LOG_COMPONENT_DEFINE("LinuxSocketFdFactory");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED(LinuxSocketFdFactory);

int sim_vprintf (const char *str, va_list args)
{
  return vprintf (str, args);
}
void *sim_malloc (unsigned long size)
{
  // XXX: use a special allocator to control memory allocated by kernel.
  return malloc (size);
}
void sim_free (void *buffer)
{
  free (buffer);
}
void *sim_memcpy (void *dst, const void *src, unsigned long size)
{
  return memcpy (dst, src, size);
}
void *sim_memset (void *dst, char value, unsigned long size)
{
  return memset (dst, value, size);
}
int sim_vasprintf (char **str, const char *fmt, va_list args)
{
  return vasprintf (str, fmt, args);
}
unsigned long sim_random (void)
{
  static UniformVariable var = UniformVariable ();
  union {
    uint8_t buffer[sizeof(unsigned long)];
    unsigned long v;
  } u;
  for (uint8_t i = 0; i < sizeof(u.buffer); i++)
    {
      u.buffer[i] = var.GetInteger(0,255);
    }
  return u.v;
}
static void 
EventTrampoline (void (*fn) (void *context), void *context, void (*pre_fn) (void),
		 EventId *event)
{
  pre_fn ();
  fn (context);
  delete event;
}
void *sim_event_schedule_ns (unsigned long ns, void (*fn) (void *context), void *context,
			     void (*pre_fn) (void))
{
  EventId *ev = new EventId ();
  EventId event = Simulator::Schedule (NanoSeconds(ns), &EventTrampoline, fn, context, pre_fn, ev);
  *ev = event;
  return ev;
}
void sim_event_cancel (void *ev)
{
  EventId *event = (EventId *)ev;
  Simulator::Remove(*event);
  delete event;
}
__u64 sim_current_ns (void)
{
  return Simulator::Now ().GetNanoSeconds ();
}

static struct SimExported *
peek_exported(TaskManager *manager)
{
  Ptr<LinuxSocketFdFactory> socketFd = manager->GetObject<LinuxSocketFdFactory> ();
  return socketFd->PeekExported ();
}


struct SimTask *sim_task_current (void)
{  
  TaskManager *manager = TaskManager::Current ();
  Task *current = manager->CurrentTask ();
  if (current->GetExtraContext () == 0)
    {
      uint32_t pid = 0;
      struct Thread *thread = (struct Thread *)current->GetContext ();
      if (thread != 0)
	{
	  pid = thread->process->pid;
	}
      current->SetExtraContext (peek_exported (manager)->task_create(current, pid));
    }
  return (struct SimTask *)current->GetExtraContext ();
}
void sim_task_wait (void)
{
  // force initialization of 'current'
  sim_task_current ();
  // now, sleep.
  TaskManager::Current ()->Sleep ();
}
int sim_task_wakeup (struct SimTask *task)
{
  TaskManager *manager = TaskManager::Current ();
  Task *other = (Task *)peek_exported(manager)->task_get_private(task);
  bool isBlocked = other->IsBlocked ();
  manager->Wakeup (other);
  return isBlocked?1:0;
}
void sim_task_yield (void)
{
  // force initialization of 'current'
  sim_task_current ();
  // now, yield.
  TaskManager::Current ()->Yield ();
}


TypeId 
LinuxSocketFdFactory::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::LinuxSocketFdFactory")
    .SetParent<SocketFdFactory> ()
    .AddConstructor<LinuxSocketFdFactory> ()
    .AddAttribute ("Library", "File to load in memory",
		   StringValue("liblinux.so"),
		   MakeStringAccessor (&LinuxSocketFdFactory::m_library),
		   MakeStringChecker ())
    ;
  return tid;
}
LinuxSocketFdFactory::LinuxSocketFdFactory ()
  : m_loader (0),
    m_exported (0),
    m_flags (0)
{}

LinuxSocketFdFactory::~LinuxSocketFdFactory ()
{
  for (uint32_t i = 0; i < m_devices.size(); i++)
    {
      // Note: we don't really destroy devices from here
      // because calling destroy requires a task context
      // m_exported->dev_destroy(m_devices[i].second);
    }
  delete m_exported;
  delete m_loader;
  m_exported = 0;
  m_loader = 0;
}

void 
LinuxSocketFdFactory::DoDispose (void)
{
  for (uint32_t i = 0; i < m_kernelTasks.size (); i++)
    {
      m_manager->Stop (m_kernelTasks[i]);
    }
  m_kernelTasks.clear ();
  m_manager = 0;
}

LinuxSocketFdFactory *
LinuxSocketFdFactory::Self (void)
{
  return PeekPointer (TaskManager::Current ()->GetObject<LinuxSocketFdFactory> ());
}

struct SimTask *
LinuxSocketFdFactory::TaskStart (void (*callback) (void *), void *context)
{
  return Self ()->DoTaskStart (callback, context);
}

struct SimTask *
LinuxSocketFdFactory::DoTaskStart (void (*callback) (void *), void *context)
{
  Task *task = m_manager->Start (callback, context, 1<<17);
  task->SetExtraContext (m_exported->task_create(task, 0));
  m_kernelTasks.push_back (task);
  return (struct SimTask *)task->GetExtraContext ();
}

void 
LinuxSocketFdFactory::DevXmit (struct SimDevice *dev, unsigned char *data, int len)
{
  return Self ()->DoDevXmit (dev, data, len);
}
void 
LinuxSocketFdFactory::DoDevXmit (struct SimDevice *dev, unsigned char *data, int len)
{
  NetDevice *nsDev = (NetDevice *)m_exported->dev_get_private (dev);
  NS_ASSERT (len >= 14);

  struct ethhdr {
    unsigned char   h_dest[6];
    unsigned char   h_source[6];
    uint16_t        h_proto;
  } *hdr = (struct ethhdr *)data;
  data += 14;
  len -= 14;
  Ptr<Packet> p = Create<Packet> (data, len);
  uint16_t protocol = ntohs (hdr->h_proto);
  Mac48Address dest;
  dest.CopyFrom (hdr->h_dest);
  nsDev->Send (p, dest, protocol);
}



struct SimExported *
LinuxSocketFdFactory::PeekExported(void)
{
  return m_exported;
}

void
LinuxSocketFdFactory::RxFromDevice (Ptr<NetDevice> device, Ptr<const Packet> p,
				    uint16_t protocol, const Address & from,
				    const Address &to, NetDevice::PacketType type)
{
  for (uint32_t i = 0; i < m_devices.size (); i++)
    {
      if (m_devices[i].first == device)
	{
	  struct SimDevice *dev = m_devices[i].second;
	  
	  struct SimDevicePacket packet = m_exported->dev_create_packet (dev, p->GetSize () + 14);
	  p->CopyData (((unsigned char *)packet.buffer) + 14, p->GetSize ());
	  struct ethhdr {
	    unsigned char   h_dest[6];
	    unsigned char   h_source[6];
	    uint16_t        h_proto;
	  } *hdr = (struct ethhdr *)packet.buffer;
	  Mac48Address realFrom = Mac48Address::ConvertFrom (from);
	  Mac48Address realTo = Mac48Address::ConvertFrom (to);
	  realFrom.CopyTo (hdr->h_source);
	  realTo.CopyTo (hdr->h_dest);
	  hdr->h_proto = ntohs (protocol);
	  m_exported->dev_rx (dev, packet);
	}
    }
}


void
LinuxSocketFdFactory::NotifyAddDevice (Ptr<NetDevice> device)
{
  int flags = 0;
  //NS_ASSERT (!device->IsPointToPoint ());
  //NS_ASSERT (device->NeedsArp ());
  NS_ASSERT (device->IsMulticast ());
  NS_ASSERT (device->IsBroadcast ());
  if (device->IsMulticast ())
    {
      flags |= SIM_DEV_MULTICAST;
    }
  if (device->IsBroadcast ())
    {
      flags |= SIM_DEV_BROADCAST;
    }
  struct SimDevice *dev = m_exported->dev_create(PeekPointer(device), (enum SimDevFlags)flags);
  // XXX: need to setup a listener to be notified of device state changes.
  // for the MTU and address at least.
  Mac48Address ad = Mac48Address::ConvertFrom (device->GetAddress ());
  uint8_t buffer[6];
  ad.CopyTo (buffer);
  m_exported->dev_set_address(dev, buffer);
  m_devices.push_back (std::make_pair(device,dev));
  Ptr<Node> node = GetObject<Node> ();
  node->RegisterProtocolHandler (MakeCallback (&LinuxSocketFdFactory::RxFromDevice, this), 
				 0, device, true);
}

void
LinuxSocketFdFactory::InitializeStackTrampoline (void *context)
{
  LinuxSocketFdFactory *self = (LinuxSocketFdFactory *)context;
  self->InitializeStack ();
  TaskManager::Current ()->Exit ();
}

void
LinuxSocketFdFactory::InitializeStack (void)
{  
  void *handle = m_loader->Load (m_library, RTLD_LOCAL);
  void *symbol = m_loader->Lookup (handle, "sim_init");
  SimInit init = (SimInit) symbol;
  if (init == 0)
    {
      NS_FATAL_ERROR ("Oops. Can't find initialization function");
    }
  m_exported = new struct SimExported ();
  struct SimImported imported;
  imported.vprintf = vprintf;
  imported.malloc = malloc;
  imported.free = free;
  imported.memcpy = sim_memcpy;
  imported.memset = sim_memset;
  imported.vasprintf = vasprintf;
  imported.random = sim_random;
  imported.event_schedule_ns = sim_event_schedule_ns;
  imported.event_cancel = sim_event_cancel;
  imported.current_ns = sim_current_ns;
  imported.task_start = &LinuxSocketFdFactory::TaskStart;
  imported.task_wait = sim_task_wait;
  imported.task_current = sim_task_current;
  imported.task_wakeup = sim_task_wakeup;
  imported.task_yield = sim_task_yield;
  imported.dev_xmit = &LinuxSocketFdFactory::DevXmit;
  init (m_exported, &imported);

  // update the linux device list with simulation device list
  Ptr<Node> node = GetObject<Node> ();
  node->RegisterDeviceAdditionListener (MakeCallback (&LinuxSocketFdFactory::NotifyAddDevice, 
						      this));  
}

void
LinuxSocketFdFactory::StartInitializationTask (void)
{
  Ptr<TaskManager> taskManager = this->GetObject<TaskManager> ();
  taskManager->Start (&LinuxSocketFdFactory::InitializeStackTrampoline, this);
}

void 
LinuxSocketFdFactory::NotifyNewAggregate (void)
{
  Ptr<Node> node = this->GetObject<Node> ();
  Ptr<LoaderFactory> loaderFactory = this->GetObject<LoaderFactory> ();
  Ptr<TaskManager> taskManager = this->GetObject<TaskManager> ();
  if (node != 0 && loaderFactory != 0 && taskManager != 0 && m_loader == 0)
    {
      m_manager = taskManager;
      m_loader = loaderFactory->Create (0, 0, 0);
      // must use ScheduleWithContext to ensure that the initialization task gets
      // a node context to be able to retrieve the task manager when it runs.
      // i.e., TaskManager::Current() needs it.
      Simulator::ScheduleWithContext (node->GetId (), Seconds (0.0),
				      &LinuxSocketFdFactory::StartInitializationTask, this);
    }
}

UnixFd *
LinuxSocketFdFactory::CreateSocket (int domain, int type, int protocol)
{
  GET_CURRENT(this << domain << type << protocol);
  struct SimSocket *socket;
  int retval = m_exported->sock_socket(domain, type, protocol, &socket);
  if (retval < 0)
    {
      return 0;
    }
  UnixFd *fd = new LinuxSocketFd (this, socket);
  return fd;
}

int 
LinuxSocketFdFactory::Close (struct SimSocket *socket)
{
  GET_CURRENT(socket);
  int retval = m_exported->sock_close (socket);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
ssize_t 
LinuxSocketFdFactory::Recvmsg(struct SimSocket *socket, struct msghdr *msg, int flags)
{
  GET_CURRENT(socket << msg << flags);
  ssize_t retval = m_exported->sock_recvmsg (socket, msg, flags | m_flags);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
ssize_t 
LinuxSocketFdFactory::Sendmsg(struct SimSocket *socket, const struct msghdr *msg, int flags)
{
  GET_CURRENT(socket << msg << flags);
  ssize_t retval = m_exported->sock_sendmsg (socket, msg, flags | m_flags);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
int 
LinuxSocketFdFactory::Getsockname(struct SimSocket *socket, struct sockaddr *name, socklen_t *namelen)
{
  GET_CURRENT(socket << name << namelen);
  int retval = m_exported->sock_getsockname (socket, name, (int*)namelen);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
int 
LinuxSocketFdFactory::Getpeername(struct SimSocket *socket, struct sockaddr *name, socklen_t *namelen)
{
  GET_CURRENT(socket << name << namelen);
  int retval = m_exported->sock_getpeername (socket, name, (int*)namelen);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
int 
LinuxSocketFdFactory::Bind (struct SimSocket *socket, const struct sockaddr *my_addr, socklen_t addrlen)
{
  GET_CURRENT(socket << my_addr << addrlen);
  int retval = m_exported->sock_bind (socket, my_addr, addrlen);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
int 
LinuxSocketFdFactory::Connect (struct SimSocket *socket, const struct sockaddr *my_addr, socklen_t addrlen)
{
  GET_CURRENT(socket << my_addr << addrlen);
  // XXX: handle O_NONBLOCK with flags.
  int retval = m_exported->sock_connect (socket, my_addr, addrlen, 0);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
int 
LinuxSocketFdFactory::Listen (struct SimSocket *socket, int backlog)
{
  GET_CURRENT(socket << backlog);
  int retval = m_exported->sock_listen (socket, backlog);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
int 
LinuxSocketFdFactory::Shutdown(struct SimSocket *socket, int how)
{
  GET_CURRENT(socket << how);
  int retval = m_exported->sock_shutdown (socket, how);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
int
LinuxSocketFdFactory::Accept (struct SimSocket *socket, struct sockaddr *my_addr, socklen_t *addrlen)
{
  GET_CURRENT(socket << my_addr << addrlen);
  struct SimSocket *newSocket;
  // XXX: handle O_NONBLOCK
  int retval = m_exported->sock_accept (socket, &newSocket, 0);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  if (my_addr != 0)
    {
      retval = m_exported->sock_getpeername (newSocket, my_addr, (int*)addrlen);
      if (retval < 0)
	{
	  current->err = -retval;
	  m_exported->sock_close (newSocket);
	  return -1;
	}
    }
  int fd = UtilsAllocateFd ();
  if (fd == -1)
    {
      m_exported->sock_close (newSocket);
      current->err = EMFILE;
      return -1;
    }

  UnixFd *unixFd = new LinuxSocketFd (this, newSocket);
  current->process->openFiles.push_back (std::make_pair (fd, unixFd));
  return fd;
}
int 
LinuxSocketFdFactory::Ioctl (struct SimSocket *socket, int request, char *argp)
{
  GET_CURRENT(socket << request << argp);
  int retval = m_exported->sock_ioctl (socket, request, argp);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
int 
LinuxSocketFdFactory::Fcntl (struct SimSocket *socket, int cmd, unsigned long arg)
{
  GET_CURRENT(socket << cmd << arg);

  if (cmd == F_SETFL && (arg & O_NONBLOCK))
    {
      m_flags |= MSG_DONTWAIT;
    }

  return 0;
}
int 
LinuxSocketFdFactory::Setsockopt (struct SimSocket *socket, int level, int optname,
				  const void *optval, socklen_t optlen)
{
  GET_CURRENT(socket << level << optname << optval << optlen);
  int retval = m_exported->sock_setsockopt (socket, level, optname, optval, optlen);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}
int 
LinuxSocketFdFactory::Getsockopt (struct SimSocket *socket, int level, int optname,
				  void *optval, socklen_t *optlen)
{
  GET_CURRENT(socket << level << optname << optval << optlen);
  int retval = m_exported->sock_getsockopt (socket, level, optname, optval, (int*)optlen);
  if (retval < 0)
    {
      current->err = -retval;
      return -1;
    }
  return retval;
}

bool 
LinuxSocketFdFactory::CheckFd (struct SimSocket *socket)
{
  GET_CURRENT(socket);

  int retval = m_exported->sim_check_socket_fd (socket);
  return retval ? true : false;
}


} // namespace ns3
